/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.decompose;

import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.kernel.extension.IElementCompositor;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;

/**
 * 
 * @author doebber
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public abstract class HierarchicElementCompositorBase<HE extends IHierarchicElementContainer>
		implements IElementCompositor<HE> {

	/** {@inheritDoc} */
	@Override
	public boolean canDecompose(EObject contained) {
		return iterateCanDecompose(contained, new CanDecomposeIterator());
	}

	/** recursive traversion of all model elements in subtree */
	private boolean iterateCanDecompose(EObject contained,
			CanDecomposeIterator cdi) {

		if (!canDecomposeSpecific(contained)) {
			return false;
		}

		if (contained instanceof IConnection) {
			IConnection c = (IConnection) contained;
			return IConnectionCompositorService.INSTANCE.canDecomposeSpecific(
					c.getSource(), c.getTarget(), c, null);
		}

		if (contained instanceof IModelElement) {
			IModelElement me = (IModelElement) contained;
			for (IModelElementSpecification spec : me.getSpecificationsList()) {
				if (cdi.contains(spec)) {
					continue;
				}
				if (!iterateCanDecompose(spec, cdi)) {
					return false;
				}
				cdi.add(spec);
			}
			// TODO CD: handle references
		}
		if (contained instanceof IHierarchicElement) {
			IHierarchicElement he = (IHierarchicElement) contained;
			for (IHierarchicElement subelement : he.getContainedElementsList()) {
				if (cdi.contains(subelement)) {
					continue;
				}
				if (!iterateCanDecompose(subelement, cdi)) {
					return false;
				}
				cdi.add(subelement);
			}
			for (IConnection conn : he.getConnectionsList()) {
				if (cdi.contains(conn)) {
					continue;
				}
				if (!iterateCanDecompose(conn, cdi)) {
					return false;
				}
				cdi.add(conn);
			}
			for (IConnector c : he.getConnectorsList()) {
				if (cdi.contains(c)) {
					continue;
				}
				if (!iterateCanDecompose(c, cdi)) {
					return false;
				}
				cdi.add(c);
			}
		}
		if (contained instanceof IConnector) {
			IConnector connector = (IConnector) contained;
			for (IConnection conn : connector.getIncomingList()) {
				if (cdi.contains(conn)) {
					continue;
				}
				if (!iterateCanDecompose(conn, cdi)) {
					return false;
				}
				cdi.add(conn);
			}
			for (IConnection conn : connector.getOutgoingList()) {
				if (cdi.contains(conn)) {
					continue;
				}
				if (!iterateCanDecompose(conn, cdi)) {
					return false;
				}
				cdi.add(conn);
			}
		}

		return true;
	}

	/** Base implementation returns true by default. Subclasses may override. */
	public boolean canDecomposeSpecific(
			@SuppressWarnings("unused") EObject contained) {
		return true;
	}

	/** Helper class to iterate over model elements in subtree to remove */
	private class CanDecomposeIterator {

		/** holds the model elements already visited */
		private HashSet<EObject> set = new HashSet<EObject>();

		/** add model element to traversed elements */
		public void add(EObject element) {
			set.add(element);
		}

		/**
		 * returns whether this model element has been inquired for
		 * decomposition ability before
		 */
		public boolean contains(EObject element) {
			return set.contains(element);
		}
	}
}
